#![allow(dead_code)]

const N: u32 = 1 << 16;

#[inline(always)]
fn part_1_by_1(mut x: u32) -> u32 {
    x &= 0x0000ffff;
    x = (x ^ (x << 8)) & 0x00ff00ff;
    x = (x ^ (x << 4)) & 0x0f0f0f0f;
    x = (x ^ (x << 2)) & 0x33333333;
    x = (x ^ (x << 1)) & 0x55555555;
    x
}

#[inline(always)]
fn part_1_by_2(mut x: u32) -> u32 {
    x &= 0x000003ff;
    x = (x ^ (x << 16)) & 0xff0000ff;
    x = (x ^ (x << 8)) & 0x0300f00f;
    x = (x ^ (x << 4)) & 0x030c30c3;
    x = (x ^ (x << 2)) & 0x09249249;
    x
}

#[inline(always)]
fn compact_1_by_1(mut x: u32) -> u32 {
    x &= 0x55555555; // x = -f-e -d-c -b-a -9-8 -7-6 -5-4 -3-2 -1-0
    x = (x ^ (x >> 1)) & 0x33333333;
    x = (x ^ (x >> 2)) & 0x0f0f0f0f;
    x = (x ^ (x >> 4)) & 0x00ff00ff;
    x = (x ^ (x >> 8)) & 0x0000ffff;
    x
}

#[inline(always)]
fn compact_1_by_2(mut x: u32) -> u32 {
    x &= 0x09249249;
    x = (x ^ (x >> 2)) & 0x030c30c3;
    x = (x ^ (x >> 4)) & 0x0300f00f;
    x = (x ^ (x >> 8)) & 0xff0000ff;
    x = (x ^ (x >> 16)) & 0x000003ff;
    x
}

/// Convert (x,y) to morton curve index.
///
/// x: The x coordinate.  Must be a positive integer no greater than 2^16-1.
/// y: The y coordinate.  Must be a positive integer no greater than 2^16-1.
///
/// Returns the morton curve index corresponding to the (x,y) coordinates given.
pub fn xy2i(x: u32, y: u32) -> u32 {
    debug_assert!(x < N);
    debug_assert!(y < N);

    part_1_by_1(x) | (part_1_by_1(y) << 1)
}

/// Convert morton curve index to (x,y).
///
/// i: The morton curve index.
///
/// Returns the (x, y) coords at the given index.
pub fn i2xy(i: u32) -> (u32, u32) {
    (compact_1_by_1(i), compact_1_by_1(i >> 1))
}

// TODO: 3D morton curves.

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn reversible() {
        let d = 54;
        let (x, y) = i2xy(d);
        let d2 = xy2i(x, y);

        assert_eq!(d, d2);
    }
}
